#
# modeler TCL script (curve.tcl)
#
# cse 457 spring 1999 project 4
#
# douglas zongker
# dougz@cs.washington.edu
#

# initialize variables holding the current animation
set g_xmax 20.0
set g_wrapflag 0
set g_curves {}
set g_curveinfo("") ""
set g_selected ""

# the set of colors available to draw animation curves in
set g_colors [list "#ff0000" "#0000ff" "#00ff00" "#008000" "#800000" "#000080" \
	           "#ffff00" "#ff00ff" "#00ffff" "#808000" "#800080" "#008080"]
set g_nextcolor 0

# add_curve is called from within controls.cpp to tell use about each
# variable that is assigned a curve.
proc add_curve {var name ymin ymax def} {
    global g_curves g_curveinfo g_colors g_nextcolor g_curveid
    global g_xmax g_wrapflag

    lappend g_curves $var

    set g_curveinfo($var,name) $name
    set g_curveinfo($var,ymin) $ymin
    set g_curveinfo($var,ymax) $ymax
    set g_curveinfo($var,def) $def
    set g_curveinfo($var,type) [lindex [get_curve_types] 0]
    set g_curveinfo($var,controls) [list ]
    set g_curveinfo($var,displayed) 0
    set g_curveinfo($var,color) [lindex $g_colors $g_nextcolor]
    set g_nextcolor [expr ($g_nextcolor+1) % [llength $g_colors]]

    evaluate_curve $g_curveinfo($var,type) $def {} 0.0 $g_xmax $g_wrapflag $var
}

# clear_curves erases the current animation
proc clear_curves {} {
    global g_curves g_curveinfo g_nextcolor g_xmax g_wrapflag g_selected
    
    set g_curves {}
    unset g_curveinfo
    set g_curveinfo("") ""
    set g_nextcolor 0
    set g_xmax 20.0
    set g_wrapflag 0

    trace vdelete g_selected w change_selection
    trace vdelete g_curvetype w change_curvetype
    
    set g_selected ""
}

# save_curves writes everything about the current animation to a file.
proc save_curves {fd} {
    global g_curves g_curveinfo g_xmax g_wrapflag

    puts $fd [list $g_xmax $g_wrapflag]
    foreach c $g_curves {
	foreach thing [list name ymin ymax def type controls] {
	    puts $fd [list $c $thing $g_curveinfo($c,$thing)]
	}
    }
}

# load_curves reads a saved animation file, and attempts to match each
# curve in the file with a curve of the current model.
proc load_curves {fd} {
    global g_curves g_curveinfo
    global g_tags g_taginfo
    global g_xmax g_wrapflag

    # the first line contains the animation's length, and whether it
    # wraps around or not.
    gets $fd line
    set g_xmax [lindex $line 0]
    set g_wrapflag [lindex $line 1]
    .cv.f.xmax configure -text [format "%.1f" $g_xmax]

    # read the file into a temporary database
    while {[gets $fd line]!=-1} {
	set g_tags([lindex $line 0]) 1
	set g_taginfo([lindex $line 0],[lindex $line 1]) [lindex $line 2]
    }

    # try matching each curve in the temp database with a curve of
    # the currently loaded model, using the names of the curves.
    foreach curve $g_curves {
	set flag 0
	foreach tag [array names g_tags] {
	    if {[string compare $g_taginfo($tag,name) $g_curveinfo($curve,name)]==0} {
		# found a match; copy everything from the temp database
		# into the real one, and redraw the curve.
		set g_curveinfo($curve,ymin) $g_taginfo($tag,ymin)
		set g_curveinfo($curve,ymax) $g_taginfo($tag,ymax)
		set g_curveinfo($curve,def) $g_taginfo($tag,def)
		set g_curveinfo($curve,type) $g_taginfo($tag,type)
		set g_curveinfo($curve,controls) $g_taginfo($tag,controls)
		reshow_curve $curve
		set flag 1
		break
	    }
	}
	if {$flag == 0} {
	    # didn't find a match for this model curve, initialize it
	    # with 0 control points.
	    set g_curveinfo($curve,controls) ""
	    reshow_curve $curve
	}
    }
}

# open_curve_editor is called to fill the .cv frame with the curve editor
proc open_curve_editor {} {
    global g_curves g_curveinfo g_selected g_xmax g_curvetype

    # .cv.f contains the edit canvas and the axis labels
    frame .cv.f
    label .cv.f.xmin -anchor w -text "0.0"
    label .cv.f.xmax -anchor e -text [format "%.1f" $g_xmax]
    label .cv.f.ymin -anchor s -text "ymin"
    label .cv.f.ymax -anchor n -text "ymax"
    canvas .cv.f.c -width 500 -height 500 -bg white

    # add mouse bindings to the edit canvas
    bind .cv.f.c <Configure> configure_canvas
    bind .cv.f.c <Motion> [list show_position %x %y]
    bind .cv.f.c <3> [list set_canvas_time %x]
    bind .cv.f.c <Button3-Motion> [list show_position %x %y]
    bind .cv.f.c <Button3-Motion> [list +set_canvas_time %x]

    # pack the axis labels in the frame with the canvas
    grid .cv.f.ymax -sticky n
    grid .cv.f.ymin -row 1 -sticky s
    grid .cv.f.c -row 0 -column 1 -rowspan 2 -columnspan 2 -sticky nsew
    grid .cv.f.xmin -row 2 -column 1 -sticky w
    grid .cv.f.xmax -row 2 -column 2 -sticky e
    grid columnconfigure .cv.f {1 2} -weight 1
    grid rowconfigure .cv.f {0 1} -weight 1
    grid columnconfigure .cv.f 0 -weight 0 -minsize 2c
    grid rowconfigure .cv.f 2 -weight 0 -minsize 1c

    # place .cv.f in the .cv frame
    pack .cv.f -padx 5 -pady 5 -fill both -expand 1

    # .cv.cont contains the row of controls at the bottom
    frame .cv.cont 

    # label for displaying the mouse position
    label .cv.cont.pos -width 10 

    # the menu of model curves
    menubutton .cv.cont.mb1 -menu .cv.cont.mb1.m -text "animation variables" -relief raised -bd 2 
    menu .cv.cont.mb1.m
    foreach item $g_curves {
	.cv.cont.mb1.m add checkbutton \
		-label $g_curveinfo($item,name) \
		-variable g_curveinfo($item,displayed) \
		-command [list toggle_enable $item] 
    }

    # display the currently selected curve
    label .cv.cont.sel -text "selected: <none>"

    # menu of curve types
    menubutton .cv.cont.mb2 -menu .cv.cont.mb2.m -text "curve settings" -relief raise -bd 2 -state disabled
    menu .cv.cont.mb2.m
    foreach item [get_curve_types] {
	.cv.cont.mb2.m add radiobutton \
		-label $item -value $item -variable g_curvetype
    }

    trace variable g_selected w change_selection
    trace variable g_curvetype w change_curvetype

    # buttons for opening dialogs
    button .cv.cont.sety -text "set y range" -command set_y_range -state disabled
    button .cv.cont.glob -text "global parameters" -command set_global

    # display number of segments in curve
    label .cv.cont.nseg -width 10
    
    # pack controls left to right in the row
    pack .cv.cont.pos .cv.cont.mb1 .cv.cont.sel .cv.cont.mb2 \
	    .cv.cont.sety .cv.cont.glob .cv.cont.nseg \
	    -side left -expand 1 -fill x 

    # add the controls to the .cv frame
    pack .cv.cont -padx 5 -pady 5 -fill x 

    configure_canvas
}

# set_global creates the (modal) global parameters dialog box.
proc set_global {} {
    global g_xmax g_wrapflag g_newwrap
    
    toplevel .glob
    wm title .glob "global parameters"

    set g_newwrap $g_wrapflag

    # allow user to change the animation length
    label .glob.l1 -text "animation length (sec):"
    entry .glob.e1 -width 10
    .glob.e1 insert end $g_xmax
    grid .glob.l1 .glob.e1 -row 0 -sticky ew -padx 3 -pady 3

    # set whether or not the animation wraps around.
    checkbutton .glob.c1 -text "wrap around" -variable g_newwrap
    grid .glob.c1 -row 1 -columnspan 2 -sticky ew -padx 3 -pady 3

    # row of buttons for accepting or cancelling the dialog.
    frame .glob.f
    button .glob.f.b1 -text "rescale" -command [list global_finish rescale]
    button .glob.f.b2 -text "clip" -command [list global_finish clip]
    button .glob.f.b3 -text "cancel" -command [list global_finish cancel]
    pack .glob.f.b1 .glob.f.b2 .glob.f.b3 -side left -padx 2 -pady 3
    grid .glob.f -row 2 -columnspan 2

    # make the dialog modal
    grab set .glob
}

# global_finish is called when one of the buttons in the global_parameter
# dialog is pushed.
proc global_finish {action} {
    global g_newwrap g_xmax g_wrapflag
    global g_curves g_curveinfo

    set newmax [.glob.e1 get]
    
    switch $action {
	rescale {
	    if {$newmax <= 0.0} {
		tk_dialog .error error "invalid length." error 0 ok
		return
	    }
	    foreach c $g_curves {
		set nc ""
		foreach p $g_curveinfo($c,controls) {
		    lappend nc [list [expr [lindex $p 0]/$g_xmax*$newmax] [lindex $p 1]]
		}
		set g_curveinfo($c,controls) $nc
	    }
	    set g_xmax $newmax
	    .cv.f.xmax config -text [format "%.1f" $newmax]
	    set g_wrapflag $g_newwrap
	    foreach c $g_curves {
		reshow_curve $c
	    }
	    destroy .glob
	}
	clip {
	    if {$newmax <= 0.0} {
		tk_dialog .error error "invalid length." error 0 ok
		return
	    }
	    foreach c $g_curves {
		set nc ""
		foreach p $g_curveinfo($c,controls) {
		    if {[lindex $p 0] <= $newmax} {
			lappend nc $p
		    }
		}
		set g_curveinfo($c,controls) $nc
	    }
	    set g_xmax $newmax
	    .cv.f.xmax config -text $newmax
	    set g_wrapflag $g_newwrap
	    foreach c $g_curves {
		reshow_curve $c
	    }
	    destroy .glob
	}
	cancel {
	    destroy .glob
	}
    }
}

# set_y_range creates the (modal) y range dialog.
proc set_y_range {} {
    global g_selected g_curveinfo

    toplevel .sety
    wm title .sety "set y range"

    # entry boxes for the new min and max
    label .sety.l1 -text "min:"
    entry .sety.e1 -width 10
    .sety.e1 insert end $g_curveinfo($g_selected,ymin) 
    label .sety.l2 -text "max:"
    entry .sety.e2 -width 10
    .sety.e2 insert end $g_curveinfo($g_selected,ymax) 
    grid .sety.l1 .sety.e1 -row 0 -sticky ew -padx 3 -pady 3
    grid .sety.l2 .sety.e2 -row 1 -sticky ew -padx 3 -pady 3

    # row of buttons for finishing the dialog
    frame .sety.f
    button .sety.f.b1 -text "rescale" -command [list set_y_finish rescale]
    button .sety.f.b2 -text "clamp" -command [list set_y_finish clamp]
    button .sety.f.b3 -text "cancel" -command [list set_y_finish cancel]
    pack .sety.f.b1 .sety.f.b2 .sety.f.b3 -side left -padx 2 -pady 3
    grid .sety.f -row 2 -columnspan 2

    # this dialog is modal
    grab set .sety
}

# set_y_finish is called when the 
proc set_y_finish {action} {
    global g_curveinfo g_selected

    set newmin [.sety.e1 get]
    set newmax [.sety.e2 get]
    
    set ok 1
    if {[catch {expr double($newmin)}] || [catch {expr double($newmax)}]} {
	set ok 0
    }
    
    switch $action {
	rescale {
	    if {$ok == 0} {
		tk_dialog .error error "invalid min/max." error 0 ok
		return
	    }
	    
	    set nc ""
	    foreach p $g_curveinfo($g_selected,controls) {
		set y [expr ([lindex $p 1]-$g_curveinfo($g_selected,ymin))/($g_curveinfo($g_selected,ymax)-$g_curveinfo($g_selected,ymin))*($newmax-$newmin)+$newmin]
		lappend nc [list [lindex $p 0] $y]
	    }
	    set g_curveinfo($g_selected,controls) $nc

	    set g_curveinfo($g_selected,ymin) $newmin
	    set g_curveinfo($g_selected,ymax) $newmax
	    
	    destroy .sety
	    reshow_curve $g_selected
	}
	clamp {
	    if {$ok == 0} {
		tk_dialog .error error "invalid min/max." error 0 ok
		return
	    }
	    set g_curveinfo($g_selected,ymin) $newmin
	    set g_curveinfo($g_selected,ymax) $newmax
	    
	    set nc ""
	    foreach p $g_curveinfo($g_selected,controls) {
		set y [lindex $p 1]
		if {$y < $newmin} {
		    set y $newmin
		} elseif {$y > $newmax} {
		    set y $newmax
		}
		lappend nc [list [lindex $p 0] $y]
	    }
	    set g_curveinfo($g_selected,controls) $nc
	    
	    destroy .sety
	    reshow_curve $g_selected
	}
	cancel {
	    destroy .sety
	}
    }
}

# change_selection is called when g_selected is written to.  it updates the
# curvetype menubutton label, and enables or disables the curvetype and
# y-range buttons as appropriate.
proc change_selection {args} {
    global g_curveinfo g_curvetype g_selected
    if {$g_selected == ""} {
	.cv.cont.mb2 configure -state disabled
	.cv.cont.sety configure -state disabled
    } else {
	.cv.cont.mb2 configure -state normal
	set g_curvetype $g_curveinfo($g_selected,type)
	.cv.cont.sety configure -state normal
    }
}

# change_curvetype is called when g_curvetype is changed.  It updates the
# currently selected curve's type and redraws it.
proc change_curvetype {args} {
    global g_curveinfo g_curvetype g_selected
    if {$g_selected != ""} {
	set g_curveinfo($g_selected,type) $g_curvetype
	reshow_curve $g_selected
    }
}
	
# set_canvas_time takes an x value in window coordinates, determines the
# corresponding time in the animation, and redraws the model at that time.
proc set_canvas_time {x} {
    global g_xmax
    set x [xunmap [.cv.f.c canvasx $x]]
    if {$x < 0.0} {set x 0.0}
    if {$x > $g_xmax} {set x $g_xmax}
    set_animation_time $x
}
    
# update_editor_time is called from within the redraw function to set the
# blue line on the edit window indicating the current time.
proc update_editor_time {time} {
    if {[winfo exists .cv.f.c]} {
	.cv.f.c delete timebar
	set x [xmap $time]
	puts "x is $x"
	.cv.f.c create line $x -10 $x 2000 -fill blue -width 1 -tags timebar
	.cv.f.c raise timebar bg
	update idletasks
    }
}

# show_position updates the position display with the current coordinates
# of the mouse.
proc show_position {x y} {
    global g_selected
    set x [format "%.2f" [xunmap $x]]
    if {[string length $g_selected]>0} {
	set y [format "%.2f" [yunmap $y $g_selected]]
	.cv.cont.pos configure -text "($x,$y)"
    } else {
	.cv.cont.pos configure -text "($x,)"
    }
}
    
# configure_canvas is called when the canvas is resized.  it draws the grid
# on the canvas, and sets the width and height variables used by the map
# and unmap functions.
proc configure_canvas {} {
    global g_curves g_curveinfo g_selected
    global g_width g_height
    
    .cv.f.c delete all

    set g_width [winfo width .cv.f.c]
    set g_height [winfo height .cv.f.c]
    
    .cv.f.c create rectangle 0 0 $g_width $g_height -fill "#e0e0e0" -tags bg
    for {set j 1} {$j < 20} {incr j} {
	set i [expr $g_height*$j/20.0]
        .cv.f.c create line -10 $i 2510 $i -fill "#c0c0c0" -tags bg
        .cv.f.c create line -10 [expr $i+1] 2510 [expr $i+1] -fill "#f0f0f0" -tags bg
	
	set i [expr $g_width*$j/20.0]
        .cv.f.c create line $i -10 $i 2510 -fill "#c0c0c0" -tags bg
        .cv.f.c create line [expr $i+1] -10 [expr $i+1] 2510 -fill "#f0f0f0" -tags bg
    }
    
    .cv.f.c bind bg <1> [list create_control %x %y]
    
    foreach item $g_curves {
	if {$g_curveinfo($item,displayed)} {
	    show_curve $item
	}
    }

    if {[string length $g_selected]>0} {
	show_controls $g_selected 1
    }
}

# toggle_enable is called when a curve is shown or hidden via the
# curve menu.
proc toggle_enable {item} {
    global g_curveinfo
    if {$g_curveinfo($item,displayed)} {
	show_curve $item
	show_controls $item
    } else {
	hide_curve $item
    }
}

# hide_curve deletes the canvas items associated with a curve
proc hide_curve {item} {
    global g_selected
    if {[string compare $g_selected $item]==0} {
	set g_selected ""
	.cv.f.ymin configure -text "ymin"
	.cv.f.ymax configure -text "ymax"
	.cv.cont.sel config -text "selected: <none>"
    }
    .cv.f.c delete $item
}

# show_curve draws a curve in the edit window, and selects it for
# editing.
proc show_curve {item} {
    global g_curveinfo g_xmax g_wrapflag

    set points [get_evaluated_curve $item]
    set coords ""
    set flag 0
    foreach p $points {
	append coords " " [xmap [lindex $p 0]] " " [ymap [lindex $p 1] $item]
	set flag 1
    }
    if {$flag} {
	set g_curveinfo($item,id) [eval ".cv.f.c create line $coords -fill $g_curveinfo($item,color) -tags $item -width 3"]
    } else {
	set g_curveinfo($item,id) 0
    }

    .cv.f.c bind $item <1> [list show_controls $item]
}

# reshow_curve updates a curve's display if it is shown, including
# control points if it is selected.
proc reshow_curve {item} {
    global g_curveinfo g_xmax g_wrapflag g_selected

    set n [evaluate_curve $g_curveinfo($item,type) $g_curveinfo($item,def) \
	    $g_curveinfo($item,controls) \
	    0 $g_xmax $g_wrapflag $item]
    .cv.cont.nseg config -text "$n pts"

    if {$g_curveinfo($item,displayed)==0} return
    
    set points [get_evaluated_curve $item]
    set coords ""
    set flag 0
    foreach p $points {
	append coords " " [xmap [lindex $p 0]] " " [ymap [lindex $p 1] $item]
	set flag 1
    }
    if {$flag} {
	if {$g_curveinfo($item,id)==0} {
	    set g_curveinfo($item,id) [eval ".cv.f.c create line $coords -fill $g_curveinfo($item,color) -tags $item -width 3"]
	} else {
	    eval ".cv.f.c coords $g_curveinfo($item,id) $coords"
	}
    } else { 
	.cv.f.c delete $g_curveinfo($item,id)
	set g_curveinfo($item,id) 0
    }

    if {$item == $g_selected} {
	show_controls $g_selected 1
    }
    
    redraw
}

# show_controls draws the control points for the currently selected curve.
proc show_controls {item {force 0}} {
    global g_curveinfo g_selected
    if {$force==0 && [string compare $g_selected $item]==0} return
    .cv.f.c delete control
    set i 0
    foreach p $g_curveinfo($item,controls) {
	set x [xmap [lindex $p 0]]
	set y [ymap [lindex $p 1] $item]
	set id1 [.cv.f.c create oval [expr $x-3] [expr $y-3] [expr $x+4] [expr $y+4] \
		-fill $g_curveinfo($item,color) -outline {} \
		-tags [list $item control]]
	set id2 [.cv.f.c create oval [expr $x-5] [expr $y-5] [expr $x+6] [expr $y+6] \
		-outline $g_curveinfo($item,color) \
		-tags [list $item control]]
	.cv.f.c bind $id1 <ButtonPress-1> [list control_down $item $i $id1 $id2]
	.cv.f.c bind $id2 <ButtonPress-1> [list control_down $item $i $id1 $id2]
	.cv.f.c bind $id1 <Control-ButtonPress-1> [list control_delete $item $i $id1 $id2]
	.cv.f.c bind $id2 <Control-ButtonPress-1> [list control_delete $item $i $id1 $id2]
	incr i
    }
    set g_selected $item
    .cv.cont.sel config -text "selected: $g_curveinfo($item,name)"
    puts "configuring labels $item $g_curveinfo($item,ymin)"
    .cv.f.ymin config -text [format "%.1f" $g_curveinfo($item,ymin)]
    .cv.f.ymax config -text [format "%.1f" $g_curveinfo($item,ymax)]
}

# create_control adds a new control point to the selected curve
proc create_control {x y} {
    global g_curveinfo g_selected

    if {[string length $g_selected]==0} return

    set i [llength $g_curveinfo($g_selected,controls)]
    lappend g_curveinfo($g_selected,controls) [list [xunmap $x] [yunmap $y $g_selected]]

    set id1 [.cv.f.c create oval [expr $x-3] [expr $y-3] [expr $x+4] [expr $y+4] \
	    -fill $g_curveinfo($g_selected,color) -outline {} \
	    -tags [list $g_selected control]]
    set id2 [.cv.f.c create oval [expr $x-5] [expr $y-5] [expr $x+6] [expr $y+6] \
	    -outline $g_curveinfo($g_selected,color) \
	    -tags [list $g_selected control]]

    control_move $g_selected $i $id1 $id2 $x $y
    
    bind .cv.f.c <Motion> [list +control_move $g_selected $i $id1 $id2 %x %y]
    bind .cv.f.c <ButtonRelease-1> [list control_up $g_selected $i $id1 $id2 %x %y]
}

# control_delete deletes a control point.
proc control_delete {item i id1 id2} {
    global g_curveinfo
    .cv.f.c delete $id1 $id2
    set g_curveinfo($item,controls) [lreplace $g_curveinfo($item,controls) $i $i]
    reshow_curve $item
    show_controls $item 1
}

# control_down is called when mouse-1 is pressed over a control, indicating
# the start of a drag.
proc control_down {item i id1 id2} {
    global g_curveinfo

    bind .cv.f.c <Motion> [list +control_move $item $i $id1 $id2 %x %y]
    bind .cv.f.c <ButtonRelease-1> [list control_up $item $i $id1 $id2 %x %y]
}

# control_move is called when the button has been pressed on a control point
# and the mouse moves.
proc control_move {item i id1 id2 x y} {
    global g_curveinfo g_width g_height

    if {$x < 0} {
	set x 0
    } elseif {$x >= $g_width} {
	set x [expr $g_width-1]
    }
    
    if {$y < 0} {
	set y 0
    } elseif {$y >= $g_height} {
	set y [expr $g_height-1]
    }
    
    .cv.f.c coords $id1 [expr $x-3] [expr $y-3] [expr $x+4] [expr $y+4]
    .cv.f.c coords $id2 [expr $x-5] [expr $y-5] [expr $x+6] [expr $y+6]

    set g_curveinfo($item,controls) [lreplace $g_curveinfo($item,controls) $i $i [list [xunmap $x] [yunmap $y $item]]]
    reshow_curve $item
}

# control_up is called when the button is released, ending a drag operation.
proc control_up {item i id1 id2 x y} {
    global g_selected

    control_move $item $i $id1 $id2 $x $y
    show_controls $item 1
    
    bind .cv.f.c <Motion> [list show_position %x %y]
    bind .cv.f.c <ButtonRelease-1> {}
}

# xmap maps an x value from animation time to canvas coordinates
proc xmap {x} {
    global g_xmax g_width
    expr double($x)/$g_xmax * $g_width
}

# xunmap maps an x value from canvas coordinates to animation time
proc xunmap {x} {
    global g_xmax g_width
    expr double($x) / $g_width * $g_xmax
}

# maps a y value from current y-range to canvas coords
proc ymap {y item} {
    global g_curveinfo g_height
    expr ($g_curveinfo($item,ymax)-double($y))/($g_curveinfo($item,ymax)-$g_curveinfo($item,ymin)) * 0.92 * $g_height + ($g_height*0.04)
}

# inverse of ymap
proc yunmap {y item} {
    global g_curveinfo g_height
    expr $g_curveinfo($item,ymax) - ((($y-0.04*$g_height)/(0.92*$g_height)) * ($g_curveinfo($item,ymax)-$g_curveinfo($item,ymin)))
}



    

    