/* -*- compile-command: "msdev modeler.dsp /make \"Modeler - Win32 Debug\"" -*-
 */

#include <modeler.h>

void linear_generate( double xmin, double xmax, int wrapflag, double def,
		      int nc, double* xc, double* yc,
		      int* np, double** xp, double** yp )
{
    double* x;
    double* y;
    int n;
    int i;
    
    // now allocate room for new points.  for the linear curve type,
    // I never need more than nc+2 points in the output.
    x = (double*)malloc( (nc+2) * sizeof(double) );
    y = (double*)malloc( (nc+2) * sizeof(double) );

    // handle the case of one or two control points.  feel free
    // to just leave this code in each of your curves.
    if ( nc == 0 || nc == 1 )
    {
	x[0] = xmin; x[1] = xmax;
	y[1] = y[0] = (nc ? yc[0] : def);
	n = 2;
	goto done;
    }

    x[0] = xmin;
    if ( wrapflag )
    {
	// if wrapping is on, I interpolate the y value at xmin and
	// xmax so that the slopes of the lines adjacent to the
	// wraparound are equal.

	y[0] = ( yc[0] * (xmax-xc[nc-1]) + yc[nc-1] * (xc[0]-xmin) ) /
	    (xc[0] - xmin + xmax - xc[nc-1]);
    }
    else
    {
	// if wrapping is off, I make the first and last segments of
	// the curve horizontal.

	y[0] = yc[0];
    }

    // a linear curve is easy; the output is mostly just the control 
    // points.

    n = 1;
    for ( i = 0; i < nc; ++i )
    {
	if ( xc[i] == x[n-1] ) continue;
	x[n] = xc[i];
	y[n] = yc[i];
	++n;
    }

    // set the endpoint based on the wrap flag.
    x[n] = xmax;
    if ( wrapflag )
	y[n] = y[0];
    else
	y[n] = y[n-1];
    ++n;

    // n should contain the number of points you are outputting.

 done:
    *np = n;
    *xp = x;
    *yp = y;
}

